/* Testbase routines for big integer manipulation */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


bigint *
bigint_create(d)
	int		d;

{
	/* Creates a bigint with value d, which should be < BIGBASE */

	bigint		*b;	/* result */

	b = (bigint *) malloc( sizeof( bigint ) );

	if ( ! b )
		fatal( "out of memory in bigint_create()" );

	bigint_set( b, d );

	return b;
}


void
bigint_set(b, d)
	bigint		*b;
	int		d;

{
	/* Set bigint b to value d, which should be |d| < BIGBASE */

	if ( d < 0 ) {
		b->bsign = negative;
		d = -d;
	} else
		b->bsign = positive;

	if ( d >= BIGBASE )
		fatal( "set value too large in bigint_set()" );

	b->blength = 1;
	b->bsignificand[0] = d;
}


void
bigint_copy(b1, b2)
	bigint		*b1, *b2;

{
	/* b2 = b1 	 */

	int             i;

	b2->bsign = b1->bsign;
	b2->blength = b1->blength;
	for (i = 0; i < b1->blength; i++)
		b2->bsignificand[i] = b1->bsignificand[i];
}


void
bigint_display(b)
	bigint		*b;
{
	/* Print a bigint on stdout. 	 */

	decimalstring	d;

	bigint_to_decimalstring( b, d );
	printf( "%c%s", (b->bsign == positive) ? '+' : '-', d );
}


void bigint_to_decimalstring( b, d )
	bigint		*b;
	decimalstring	d;
{
	/* Converts b to its decimalstring equivalent. */

	bigint		*power10, quotient, remainder;
	EXPONENT	log_10_b, num_digits, num_digits_extracted;
	int		compare;
	static int	bigbase_log_10 = -1;

	if ( b->blength == 1 && b->bsignificand[0] <= 1 ) {
		/* These special cases give us headaches because of their
		 * non-positive logs.  Take care of them directly.
		 */
		d[0] = '0' + b->bsignificand[0];	/*###*/
		d[1] = 0;
		return;
	}

	if ( bigbase_log_10 < 0 ) {	/* Need to compute bigbase_log_10. */
		WORK	power_of_ten;

		for ( power_of_ten = 1.0; power_of_ten < (WORK) BIGBASE;
		      power_of_ten *= 10.0 )
			++bigbase_log_10;

		/* bigbase_log_10 is now floor(log(BIGBASE,10)), i.e.,
		 * greatest integer k such that 10^k < BIGBASE.
		 */
	}

	/* Find log(b,10).  Then floor(log(b,10)) + 1 is exactly how
	 * many decimal digits it takes to represent b.
	 */

	/* First form an estimate, then correct it as needed. */
	log_10_b = (WORK) (b->blength * BIGBASE_BITS) / LOG_10_TO_BASE_2;

	power10 = ten_to_power( log_10_b );
	compare = bigint_compare( power10, b );
	if ( compare > 0 ) {
		/* Estimate is too high. */
		do
			power10 = ten_to_power( --log_10_b );
		while ( bigint_compare( power10, b ) > 0 );

	} else if ( compare < 0 ) {
		/* Estimate is perhaps too low. */
		do
			power10 = ten_to_power( ++log_10_b );
		while ( bigint_compare( power10, b ) < 0 );
		--log_10_b;

	} /* else estimate is exact! */

	num_digits = log_10_b + 1;

	/* The most decimal digits that can fit into a single bigint is
	 * bigbase_log_10, so we can't extract more than that many
	 * in one shot.
	 */
	num_digits_extracted = bigbase_log_10;

	/* But in any case we can't extract more digits than are in the
	 * number.
	 */
	if ( num_digits_extracted > num_digits )
		num_digits_extracted = num_digits;

	/* We now can extract num_digits_extracted digits by dividing b by
	 * 10^(num_digits-num_digits_extracted).
	 */
	if ( num_digits > num_digits_extracted ) {
		bigint_divide( b,
				ten_to_power( num_digits-num_digits_extracted ),
				&quotient, &remainder );
	} else {
		bigint_copy( b, &quotient );
		bigint_set( &remainder, 0 );
	}

	if ( quotient.blength > 1 )
		fatal( "math error in bigint_to_decimalstring()" );

	/* Okay, quotient now represents the top bits in the decimal string. */
	sprintf( d, "%lu", quotient.bsignificand[0] );

	/* Repeat this process on the remaining digits till we've
	 * exhausted them.
	 */
	while ( (num_digits -= num_digits_extracted) > 0 ) {
		num_digits_extracted = bigbase_log_10;
		if ( num_digits_extracted > num_digits )
			num_digits_extracted = num_digits;

		if ( num_digits > num_digits_extracted ) {
			bigint_divide( &remainder,
				ten_to_power( num_digits-num_digits_extracted ),
					&quotient, &remainder );
		} else {
			bigint_copy( &remainder, &quotient );
			bigint_set( &remainder, 0 );
		}

		if ( quotient.blength > 1 )
			fatal( "math error in bigint_to_decimalstring()" );

		sprintf( d + strlen(d), "%0*lu", (int) num_digits_extracted,
			quotient.bsignificand[0] );
	}
}


void
bigint_grow( b, new_size )
	bigint		*b;
	int		new_size;

{	/* increase the number of digits used by a bigint to new_size */
	b->blength = new_size;
	if (b->blength >= BIGINTSIZE) {	/* Overflow! */
		fatal("BIGINTSIZE TOO SMALL in bigint_grow()");
	}
}


int
bigint_is_zero( b )
	bigint		*b;
{
	/* Returns true if a (normalized) bigint is 0, false otherwise */
	return b->blength == 1 && b->bsignificand[0] == 0;
}


void
bigint_normalize( b )
	bigint		*b;
{
	/* Reduce the length of b to reflect the first non-zero digit,
	 * or 1 if all digits are zero.
	 */

	int		digit;

	for ( digit = b->blength - 1; digit > 0 && b->bsignificand[digit] == 0;
	      --digit )
		;

	if ( digit == 0 && b->bsignificand[digit] == 0 )
		/* Make sure we don't introduce any -0's */
		b->bsign = positive;

	b->blength = digit + 1;
}


BIGBASE_TYPE
bigint_normalize_digit( b, d )
	bigint		*b;
	int		d;
{
	/* If digit b[d] is not within the bounds for a bigint digit,
	 * bring it within bounds and return the excess as a carry.
	 * Otherwise, return 0 as a carry.
	 */

	BIGBASE_TYPE	carry;

	carry = b->bsignificand[d] >> BIGBASE_BITS;
	b->bsignificand[d] = b->bsignificand[d] % BIGBASE;

	return carry;
}


int
bigint_compare( b1, b2 )
	bigint		*b1, *b2;
{
	/* Compare bigint b1 with bigint b2, returning 1 if |b1| > |b2|,
	 * 0 if |b1| == |b2|, and -1 if |b1| < |b2|.
	 */

	bigint_normalize( b1 );
	bigint_normalize( b2 );

	if ( b1->blength > b2->blength )
		return 1;

	else if ( b1->blength < b2->blength )
		return -1;

	else {
		int		i;

		for ( i = b1->blength - 1; i >= 0; --i ) {
			if ( b1->bsignificand[i] > b2->bsignificand[i] )
				return 1;

			else if ( b1->bsignificand[i] < b2->bsignificand[i] )
				return -1;
		}
	}

	return 0;
}

void
bigint_multiply_digits(b1, b2, d1, d2, b3)
	bigint		*b1, *b2, *b3;
	int		d1, d2;
{
	/* Compute b3[d1+d2] = b1[d1] * b2[d2] (with possible carry).
	 * The computation is done by splitting each digit into halves,
	 * multiplying the halves, and collected a "result" and a "carry".
	 * We decompose b1[d1] into (a << half_size) + b, and b2[d2]
	 * into (c << half_size) + d.
	 */
	BIGBASE_TYPE	result, carry;
	BIGBASE_TYPE	a, b, c, d;

	BIGBASE_TYPE	ac, bc, ad, bd;	/* intermediate products */
	int		result_digit;	/* where in b3 to put result */


	a = BIGINT_EXTRACT_UPPER(b1->bsignificand[d1]);
	c = BIGINT_EXTRACT_UPPER(b2->bsignificand[d2]);

	b = BIGINT_EXTRACT_LOWER(b1->bsignificand[d1]);
	d = BIGINT_EXTRACT_LOWER(b2->bsignificand[d2]);

	ac = a * c;
	bc = b * c;

	ad = a * d;
	bd = b * d;

	result = bd + (BIGINT_EXTRACT_LOWER(bc) << BIGBASE_HALF_BITS) +
			(BIGINT_EXTRACT_LOWER(ad) << BIGBASE_HALF_BITS);
	carry = ac + BIGINT_EXTRACT_UPPER(bc) + BIGINT_EXTRACT_UPPER(ad);

	/* Now add the result into b3 and adjust the carry. */
	result_digit = d1 + d2;
	b3->bsignificand[result_digit] += result;

	carry += bigint_normalize_digit( b3, result_digit );

	/* Propagate carry. */
	while ( carry > 0 ) {
		if ( ++result_digit >= b3->blength )
			fatal(
			"internal inconsistency in bigint_multiply_digits");

		b3->bsignificand[result_digit] += carry;
		carry = bigint_normalize_digit( b3, result_digit );
	}
}


void
bigint_times_half_digit_shifted( b1, hd, shift, b2 )
	bigint		*b1, *b2;
	BIGBASE_TYPE	hd;
	int		shift;
{
	/* Computes b2 = (b1 * hd) << (BIGBASE_HALF_BITS * shift),
	 * where 0 <= hd < (1 << BIGBASE_HALF_BITS) and "shift" is the
	 * (non-negative) number of half-digits by which to shift up
	 * the result.
	 */
	int		id;
	int		target;	/* digit where to put current result */
	BIGBASE_TYPE	p1, p2, p3, carry, lower_half_digit, upper_half_digit;
	int		shift_is_even;

	shift_is_even = (shift % 2 == 0);

	for ( target = 0; target < shift / 2; ++target )
		b2->bsignificand[target] = 0;
	b2->bsignificand[target] = 0;

	carry = 0;
	for ( id = 0, target = shift / 2; id < b1->blength; ++id, ++target ) {
		lower_half_digit = BIGINT_EXTRACT_LOWER(b1->bsignificand[id]);
		upper_half_digit = BIGINT_EXTRACT_UPPER(b1->bsignificand[id]);

		p1 = lower_half_digit * hd;
		p2 = upper_half_digit * hd;

		p3 = p1 + carry;
		p3 += BIGINT_EXTRACT_LOWER(p2) << BIGBASE_HALF_BITS;

		if ( shift_is_even )
			b2->bsignificand[target] = (p3 & (BIGBASE - 1));
		else
			{
			b2->bsignificand[target] |=
				BIGINT_EXTRACT_LOWER(p3) << BIGBASE_HALF_BITS;
			b2->bsignificand[target + 1] = BIGINT_EXTRACT_UPPER(p3);
			}

		carry = BIGINT_EXTRACT_UPPER(p2);

		if ( p3 >= BIGBASE )
			carry += p3 >> BIGBASE_BITS;
	}

	if ( shift_is_even )
		b2->bsignificand[target] = carry;

	else
		b2->bsignificand[target] |= carry << BIGBASE_HALF_BITS;

	b2->blength = target + 1;
	b2->bsign = b1->bsign;
	bigint_normalize( b2 );
}


void
bigint_multiply(b1, b2, b3)
	bigint		*b1, *b2, *b3;
{
	/* b3 = b1 * b2; b3 must be a distinct bigint pointer from b1 and b2. */

	int		result_size;
	int		d1, d2, d3;

	if ( b3 == b1 || b3 == b2 )
		fatal( "non-distinct pointer passed to bigint_multiply()" );

	/* Initialize result to all zeroes, large enough to hold
	 * final product.
	 */
	result_size = b1->blength + b2->blength;
	bigint_grow( b3, result_size );

	for ( d3 = 0; d3 < result_size; ++d3 )
		b3->bsignificand[d3] = 0;

	for ( d1 = 0; d1 < b1->blength; ++d1 )
		for ( d2 = 0; d2 < b2->blength; ++d2 )
			bigint_multiply_digits( b1, b2, d1, d2, b3 );

	bigint_normalize( b3 );

	if ( b1->bsign == b2->bsign )
		b3->bsign = positive;
	else
		b3->bsign = negative;
}


void
bigint_divide(b1, b2, b3, b4)
	bigint		*b1, *b2, *b3, *b4;
{
	/* Divide b1 by b2, putting quotient in b3 and remainder in b4.
	 * b1, b2, b3, and b4 need not be distinct bigint pointers
	 * (except b3 and b4 should be distinct).
	 */

	bigint		v_storage, v_prime_storage;
	bigint		*u;				/* u is dividend */
	bigint		*v = &v_storage;		/* v is divisor */
	bigint		*q;				/* q is quotient */
	bigint		*v_prime = &v_prime_storage;	/* temporary */
	int		len_u, len_v, j;
	BIGBASE_TYPE	d;
	EXPONENT	normalization_shift;

	/* We follow fairly closely Algorithm 4.3.1D from Knuth Vol. 2, 2nd
	 * edition, pp. 257-258.  Note that we divide our bigint "digit"'s
	 * into two halves, so we can be assured that we can use hardware
	 * division to compute the quotient of a value comprising two halves
	 * divided by a value comprising of one half.  Therefore, all of
	 * the computations are carried out with radix = BIGBASE/2.
	 *
	 * Subsequently, when we refer to a DIGIT we will mean a digit
	 * in radix BIGBASE/2.
	 * 
	 */

	q = b3;
	u = b4;

	bigint_copy( b1, u );
	bigint_copy( b2, v );

	u->bsign = positive;
	v->bsign = positive;

	if ( bigint_is_zero( v ) )
		fatal( "divide by zero in bigint_divide()" );

	len_u = u->blength * 2;
	len_v = v->blength * 2;

	if ( len_u <= len_v && bigint_compare( u, v ) < 0 ) {
		/* Easy - quotient is 0 and remainder is u. */
		bigint_copy( &big_zero, q );
		return;
	}

	if ( EXTRACT_HALF_DIGIT(u, len_u - 1) == 0 )
		--len_u;

	if ( EXTRACT_HALF_DIGIT(v, len_v - 1) == 0 )
		--len_v;

	/* Step D1: Normalize u and v so that the top DIGIT of
	 * v is >= HALF_RADIX/2.
	 */

	d = EXTRACT_HALF_DIGIT(v, len_v - 1);
	normalization_shift = 0;

	while ( d < HALF_RADIX / 2 ) {
		d <<= 1;
		++normalization_shift;
	}

	SET_HALF_DIGIT(u, len_u, 0);
	++len_u;

	if ( normalization_shift != 0 ) {
		bigint_times_2_to_power( u, normalization_shift );
		bigint_times_2_to_power( v, normalization_shift );
	}


	q->blength = (len_u - len_v) / 2 + 1;
	q->bsignificand[q->blength - 1] = 0;

	/* Steps D2 and D7: compute DIGITs of quotient from most-to-least
	 * significant
	 */

	for ( j = len_u - 1; j >= len_v; --j ) {
		BIGBASE_TYPE	q_hat, uj, uj1, uj2, v1, v2;

		uj2 = 0;
		v2 = 0;

		uj = EXTRACT_HALF_DIGIT(u, j);
		uj1 = EXTRACT_HALF_DIGIT(u, j - 1);

		if ( j > 1 )
			uj2 = EXTRACT_HALF_DIGIT( u, j - 2 );

		v1 = EXTRACT_HALF_DIGIT( v, len_v - 1 );

		if ( len_v > 1 )
			v2 = EXTRACT_HALF_DIGIT( v, len_v - 2 );


		/* D3: form estimate, q_hat, of quotient digit */
		if ( uj == v1 )
			q_hat = HALF_RADIX - 1;
		else
			q_hat = ((uj * HALF_RADIX) + uj1) / v1;

		/* Knuth tells us that the following black magic test is
		 * a cheap way to eliminate almost all overestimates of q_hat.
		 */
		while ( v2*q_hat >
			(uj*HALF_RADIX + uj1 - q_hat*v1) * HALF_RADIX + uj2 )
			/* Oops, estimate is too high */
			--q_hat;


		/* D4: multiply-and-subtract */
		bigint_times_half_digit_shifted( v, q_hat, j - len_v, v_prime );

		/* D5, D6: test whether the guess is too large and if
		 * so correct it.
		 */
		if ( bigint_compare( v_prime, u ) > 0 ) {
			/* Oops, q_hat was too ambitious. */
			bigint		v_double_prime;

			/* Knuth promises us that q_hat is at most one
			 * too high
			 */
			--q_hat;
			bigint_times_half_digit_shifted( v, (BIGBASE_TYPE) 1,
				j - len_v, &v_double_prime );

			bigint_add( u, &v_double_prime, u );
		}

		bigint_subtract( u, v_prime, u );

		if ( u->bsign != positive )
			fatal( "algorithm flaw detected in bigint_divide()" );

		/* Set digit in quotient ... */
		SET_HALF_DIGIT( q, j - len_v, q_hat );
	}

	/* Okay, now the quotient is in q.  Step D8: Unnormalize remainder. */
	if ( normalization_shift != 0 )
		bigint_times_2_to_power( u, -normalization_shift );

	bigint_normalize( q );
	bigint_normalize( u );

	if ( b1->bsign == b2->bsign )
		q->bsign = positive;
	else
		q->bsign = negative;
}


void
bigint_times_ten(b, carry)
	bigint         *b;
	BIGBASE_TYPE    carry;
{				/* b = b * 10 + carry */

	int             id;
	BIGBASE_TYPE    a;	/* accumulator */

	/* Note that this routine is beaten on fairly heavily by
	 * decimalstring_to_unpacked() so we want it to go as fast
	 * as possible.
	 */

	for (id = 0; id < b->blength; ++id) {
		/* Multiply each digit by 10, with carry in and carry out. */
		a = b->bsignificand[id] * 10 + carry;
		carry = a >> BIGBASE_BITS;
		b->bsignificand[id] = a % BIGBASE;
	}
	if (carry != 0) {
		b->blength++;
		b->bsignificand[id] = carry;
	}
}


void
two_to_power( e, b )
	EXPONENT	e;
	bigint		*b;
{				/* Returns b = 2^e for non-negative e */
	int		id;

	for ( id = 0; e >= BIGBASE_BITS; ++id ) {
		b->bsignificand[id] = 0;
		e -= BIGBASE_BITS;
	}

	/* e is now in the range [0 .. BIGBASE_BITS - 1] */
	b->bsignificand[id] = 1 << e;
	b->blength = id + 1;
}


bigint *
ten_to_power( e )
	EXPONENT	e;
{				/* Returns a bigint = 10^e for non-negative e */

	static bigint	**known_powers;	/* powers we've already computed */
	static int	size_known_powers = 0;	/* size of known_powers array */
	int		power;

	if ( size_known_powers <= e ) {
		/* Increase size of known_powers */
		int		prev_size;
		unsigned	alloc_size;

		if ( ! known_powers ) {
			size_known_powers = 100;
			alloc_size = sizeof( bigint * ) * size_known_powers;
			known_powers = (bigint **) malloc( alloc_size );
			
			known_powers[0] = &big_one;
			known_powers[1] = bigint_create( 10 );
			known_powers[2] = bigint_create( 100 );

			prev_size = 3;

		} else	{
			prev_size = size_known_powers;
			while ( size_known_powers <= e )
				size_known_powers *= 2;

			alloc_size = sizeof( bigint * ) * size_known_powers;
			known_powers = (bigint **)
				realloc( (ALLOC_TYPE) known_powers,
					 alloc_size );
		}

		if ( ! known_powers )
			fatal( "out of memory in ten_to_power()" );

		for ( power = prev_size; power < size_known_powers; ++power )
			known_powers[power] = (bigint *) 0;
	}

	if ( ! known_powers[e] ) {
		/* Need to compute this power.  Let f and g be those numbers
		 * closest to e/2 (if e is even, then f = g = e/2, otherwise
		 * f = g-1 = e div 2).  Find 10^f and 10^g and then compute
		 * 10^e by multiplying those two results.
		 */
		EXPONENT	f, g;
		bigint		*ten_to_f, *ten_to_g, *ten_to_e;

		f = g = e / 2;
		if ( e % 2 )
			++g;

		ten_to_f = ten_to_power( f );
		ten_to_g = ten_to_power( g );

		ten_to_e = bigint_create( 0 );

		if ( ! ten_to_e )
			fatal( "out of memory in ten_to_power()" );

		bigint_multiply( ten_to_f, ten_to_g, ten_to_e );

		known_powers[e] = ten_to_e;
	}

	return known_powers[e];
}

void
bigint_times_2_to_power(b, e)
	bigint		*b;
	EXPONENT	e;

{
	/* b = b * 2**e; if e < 0, we truncate bits shifted off the bottom */

	/* Note: we do the multiplication by shifting, so it's essential
	 * that the bigint radix be a power of two ...
	 */

	/* If e >= 0, we shift from the most-significant digit down to the
	 * least, otherwise in the opposite order.
	 *
	 * Unless the shift happens to be a multiple of the bigint radix,
	 * each digit when shifted will have its most-significant portion
	 * (i.e., the "top bits") shifted into one destination digit and
	 * the least-significant portion shifted one digit less.
	 *
	 * For example, suppose the bigint radix is 2^15 and we want to
	 * shift by 36.  Then we will take the top 6 bits of each digit d
	 * and deposit them as the bottom 6 bits of digit d+3, and take
	 * the lower 9 bits and deposit them as the upper 9 bits of digit d+2.
	 *
	 * In the code that follows:
	 *
	 *	shift_digits is the number of digits ahead (or behind) that
	 *		the upper bits will be placed in (3 in our example,
	 *		since the top bits are moved from digit d to digit d+3);
	 *
	 *	top_bits is the number of bits which will be shifted into
	 *		digit d+shift_digits (6 in our example)
	 *
	 *	top_mask is a mask for extracting the top bits
	 *		(0x3f << 9, in our example)
	 *
	 *	bottom_bits is the number of bits which will be shifted
	 *		into digit d+shift_digits-1 (9, in our example)
	 *
	 *	bottom_mask is a mask for extracting those bits
	 *		(0x1ff, in our example)
	 */

	int		d;
	unsigned int	shift_digits;
	BIGBASE_TYPE	top_bits, top_mask, bottom_bits, bottom_mask;

	if ( e == 0 )
		return;

	d = b->blength - 1;	/* most significant digit at present */

	if ( e > 0 ) {
		shift_digits = e / BIGBASE_BITS + 1;
		top_bits = e % BIGBASE_BITS;
		bottom_bits = BIGBASE_BITS - top_bits;
	} else {
		shift_digits = (-e) / BIGBASE_BITS;
		bottom_bits = (-e) % BIGBASE_BITS;
		top_bits = BIGBASE_BITS - bottom_bits;
	}

	top_mask = ((1 << top_bits) - 1) << bottom_bits;
	bottom_mask = (1 << bottom_bits) - 1;

	if ( e > 0 ) {	/* Shifting up */
		bigint_grow( b, (int) (b->blength + shift_digits) );

		/* Make sure new, final (most-significant) digit is 0. */
		b->bsignificand[b->blength - 1] = 0;

		while ( d >= 0 ) {
			b->bsignificand[d + shift_digits] |=
				(b->bsignificand[d] & top_mask) >> bottom_bits;

			b->bsignificand[d + shift_digits - 1] =
				(b->bsignificand[d] & bottom_mask) << top_bits;

			--d;
		}

		/* Now set to zero any digits which have been wholly vacated.
		 * The previously least-significant digit has now been shifted
		 * into digits shift_digits and shift_digits - 1.  Therefore
		 * digits 0 .. shift_digits - 2 have been vacated.
		 */
		for ( d = 0; d < shift_digits - 1; ++d )
			b->bsignificand[d] = 0;

	} else {	/* Shifting down */
		int		i, offset;

		for (i = 0; i <= d; ++i) {
			offset = i - shift_digits;

			if ( offset > 0 )
				b->bsignificand[offset - 1] |=
					(b->bsignificand[i] & bottom_mask) <<
					top_bits;

			if ( offset >= 0 )
				b->bsignificand[offset] =
					(b->bsignificand[i] & top_mask) >>
					bottom_bits;
		}

		b->blength = offset + 1;
	}

	bigint_normalize( b );
}

void
bigint_times_power(b, base, e)
	bigint		*b;
	unsigned	base;
	EXPONENT	e;

{
	/* b = b * base**e	for e >= 0	 */

	if ( e == 0 )
		return;	/* speed hack */

	if ( base == 2 )
		bigint_times_2_to_power( b, e );

	else if ( base == 10 ) {
		bigint		result;

		result.bsign = b->bsign;
		bigint_multiply( b, ten_to_power( e ), &result );
		bigint_copy( &result, b );

	} else
		fatal( "bad base in bigint_times_power()" );
}

void
bigint_add(b1, b2, b3)
	bigint		*b1, *b2, *b3;

{
	/*
	 * Compute the sum b3 = b1 + b2.  b3 needn't be a distinct bigint
	 * pointer from b1 or b2.
	 */

	int		id;
	BIGBASE_TYPE	carry;
	long		a;

	if ( b1->bsign != b2->bsign ) {
		/* This is really a subtraction */
		if ( bigint_compare( b1, b2 ) >= 0 ) {
			/* Largest is b1; effect addition by subtracting
			 * -b2 from b1.
			 */
			enum fsign	b2sign;

			b2sign = b2->bsign;
			b2->bsign = b1->bsign;
			bigint_subtract( b1, b2, b3 );

			if ( b2 != b3 )
				b2->bsign = b2sign;

		} else {
			/* Largest is b2; effect addition by subtracting
			 * -b1 from b2.
			 */
			enum fsign	b1sign;

			b1sign = b1->bsign;
			b1->bsign = b2->bsign;
			bigint_subtract( b2, b1, b3 );

			if ( b1 != b3 )
				b1->bsign = b1sign;
		}

		return;
	}

	b3->bsign = b1->bsign;

	carry = 0;
	for (id = 0; (id < b1->blength) && (id < b2->blength); ++id) {
		a = b1->bsignificand[id] + b2->bsignificand[id] + carry;
		carry = a >> BIGBASE_BITS;
		a &= (BIGBASE - 1);
		b3->bsignificand[id] = a;
	}
	while (id < b1->blength) {
		a = b1->bsignificand[id] + carry;
		carry = a >> BIGBASE_BITS;
		a &= (BIGBASE - 1);
		b3->bsignificand[id] = a;
		id++;
	}
	while (id < b2->blength) {
		a = b2->bsignificand[id] + carry;
		carry = a >> BIGBASE_BITS;
		a &= (BIGBASE - 1);
		b3->bsignificand[id] = a;
		id++;
	}
	b3->bsignificand[id] = carry;
	b3->blength = id + 1;
	bigint_normalize( b3 );
}

void
bigint_subtract(b1, b2, b3)
	bigint		*b1, *b2, *b3;

{
	/* Compute the difference b3 = b1 - b2.  b1, b2, and b3 need not
	 * be distinct.
	 */

	int             id;
	long            a, borrow;
	int             reverse;

	if ( b1->bsign != b2->bsign ) {
		/* This is really an addition */
		enum fsign	b2sign;

		b2sign = b2->bsign;
		b2->bsign = b1->bsign;
		bigint_add( b1, b2, b3 );

		if ( b2 != b3 )
			b2->bsign = b2sign;

		return;
	}

	if ( bigint_compare( b1, b2 ) < 0 ) {
		/* Oops - the subtraction will cause a sign-flip.  Instead,
		 * compute b2-b1 and flip the sign on the result.
		 */
		bigint		*temp;

		temp = b1;
		b1 = b2;
		b2 = temp;

		reverse = 1;

	} else
		reverse = 0;

	borrow = 0;
	for (id = 0; (id < b1->blength) && (id < b2->blength); ++id) {
		a = b1->bsignificand[id] - b2->bsignificand[id] - borrow;
		if (a < 0) {
			borrow = 1;
			b3->bsignificand[id] = a + BIGBASE;
		} else {	/* a >= 0 */
			borrow = 0;
			b3->bsignificand[id] = a;
		}
	}
	while (id < b1->blength) {
		a = b1->bsignificand[id] - borrow;
		if (a < 0) {
			borrow = 1;
			b3->bsignificand[id] = a + BIGBASE;
		} else {	/* a >= 0 */
			borrow = 0;
			b3->bsignificand[id] = a;
		}
		id++;
	}
	while (id < b2->blength) {
		a = -b2->bsignificand[id] - borrow;
		if (a < 0) {
			borrow = 1;
			b3->bsignificand[id] = a + BIGBASE;
		} else {	/* a >= 0 */
			borrow = 0;
			b3->bsignificand[id] = a;
		}
		id++;
	}
	if (borrow != 0) {	/* Sign reversal occurred - previous result
				 * invalid. */
		fatal( "unanticipated borrow in bigint_subtract() ");
	}
	b3->blength = id;

	if ( reverse )
		b3->bsign = (b1->bsign == positive ? negative : positive);
	else
		b3->bsign = b1->bsign;

	bigint_normalize( b3 );
}

void
bigint_remainder(b1, b2, pntimes)
	bigint		*b1, *b2;
	int		*pntimes;

{
	/*
	 * Let b0 be the original value of b1.  Compute pntimes = n to
	 * minimize |b1| where b1 = b0 - n * b2.  But limit |n| to 4;
	 * if |n| > 4 we don't need to compute ulps errors accurately.
	 * The altered value of b1 is returned.
	 */

	int             ntimes;
	enum fsign	orig_b1_sign, orig_b2_sign;

	orig_b1_sign = b1->bsign;
	orig_b2_sign = b2->bsign;

	/* We just want to work with the magnitudes of the numbers. */
	b2->bsign = b1->bsign = positive;

	for (ntimes = 1; ntimes <= 4; ++ntimes ) {
		bigint_subtract( b1, b2, b1 );

		if ( bigint_is_zero( b1 ) ) {
			/* Exact match - all done. */
			break;
		}

		if ( b1->bsign == negative ) {
			/* Okay, we've now subtracted off enough b2's that
			 * |b0| < ntimes*|b2|.  Find the lesser of
			 * |b0 - ntimes*|b2|| and |b0 - (ntimes-1)*|b2||
			 * and that's the one we want to keep.
			 */

			bigint		b1_plus_b2;

			bigint_add( b1, b2, &b1_plus_b2 );

			if ( bigint_compare( b1, &b1_plus_b2 ) >= 0 ) {
				/* |b0 - ntimes*|b2|| >= |b0 - (ntimes-1)*|b2||;
				 * pick n = ntimes-1, adjust b1 accordingly.
				 */
				bigint_copy( &b1_plus_b2, b1 );
				--ntimes;
			}
			break;
		}
	}

	if ( ntimes > 4 )
		ntimes = 4;

	if ( b1->bsign == positive )
		b1->bsign = orig_b1_sign;
	else
		b1->bsign = (orig_b1_sign == positive) ? negative : positive;

	b2->bsign = orig_b2_sign;

	if ( orig_b1_sign == orig_b2_sign )
		*pntimes = ntimes;
	else
		*pntimes = -ntimes;
}

EXPONENT
bigint_num_bits( b )
	bigint		*b;
{
	/* Returns the number of bits needed to hold the value of b. */

	EXPONENT	num_bits;
	BIGBASE_TYPE	top_bits;

	num_bits = (b->blength - 1) * BIGBASE_BITS;
	top_bits = b->bsignificand[b->blength - 1];

	while ( top_bits ) {
		++num_bits;
		top_bits >>= 1;
	}

	return num_bits;
}
