/* Testbase routines for modulus-minimization */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


static bigint	*a, *b;
static bigint	last_a, last_b;

static int	size_convergents_arrays = 0;
static int	num_convergents;
static int	num_computed_convergents;

static int	last_odd_conv, last_even_conv;

static bigint	*conv_denom = 0;
static bigint	*mod_value = 0;	/* value of a*conv_denom % b */


void
init_for_convergents( new_a, new_b )
	bigint		*new_a, *new_b;
{
	/* Initialize for computing (denominators of) convergents of a/b;
	 * assumes 0 <= a <= b, gcd(a,b) == 0.
	 */
	unsigned	convergents_size;

	if ( bigint_compare( new_a, new_b ) >= 0 )
		fatal( "a >= b in init_for_convergents()" );

	a = new_a;
	b = new_b;

	bigint_copy( a, &last_a );
	bigint_copy( b, &last_b );

	/* Compute the maximum number of convergents a/b might have.  Note
	 * that this is bounded by 1.5 * lg b; we also leave room for our
	 * first two denominators, 0 and 1 ...
	 */

	num_convergents = 3 * (b->blength * BIGBASE_BITS) / 2;
	num_convergents += 4;	/* add in room for 0 and 1, plus slop */

	if ( num_convergents > size_convergents_arrays ) {
		/* Existing arrays aren't big enough. */
		convergents_size = sizeof( bigint ) * num_convergents;

		if ( conv_denom ) {
			free( (ALLOC_TYPE) conv_denom );
			free( (ALLOC_TYPE) mod_value );
		}

		conv_denom = (bigint *) malloc( convergents_size );
		mod_value = (bigint *) malloc( convergents_size );

		size_convergents_arrays = num_convergents;
	}

	if ( ! conv_denom || ! mod_value )
		fatal( "out of memory in init_for_convergents()" );

	bigint_set( &conv_denom[0], 0 );
	bigint_set( &conv_denom[1], 1 );

	bigint_set( &mod_value[0], 0 );
	bigint_copy( a, &mod_value[1] );

	num_computed_convergents = 2;
}


void
compute_modulus( i, modulus )
	bigint		*i, *modulus;
{
	/* Computes a*i % b. */

	bigint		temp, quotient;

	bigint_multiply( a, i, &temp );
	bigint_divide( &temp, b, &quotient, modulus );
}


void
next_convergent()
{
	/* Compute the next convergent after the last one we've done.
	 * Returns the index of the convergent.
	 */

	bigint		quotient, remainder;

	if ( bigint_is_zero( &last_b ) ||
	     (num_computed_convergents >= num_convergents) )
		fatal( "next_convergent() called beyond last convergent" );

	/* Try to save a divide if possible ... */

	bigint_subtract( &last_b, &last_a, &remainder );

	if ( bigint_compare( &remainder, &last_a ) >= 0 )
		bigint_divide( &last_b, &last_a, &quotient, &remainder );
	else
		bigint_set( &quotient, 1 );

	/* New convergent denominator is
	 *	conv_denom[0] = conv_denom[-2] + quotient * conv_denom[-1]
	 */
	bigint_multiply( &conv_denom[num_computed_convergents - 1], &quotient,
			 &conv_denom[num_computed_convergents] );
	bigint_add( &conv_denom[num_computed_convergents],
		    &conv_denom[num_computed_convergents - 2],
		    &conv_denom[num_computed_convergents] );

	bigint_copy( &last_a, &last_b );
	bigint_copy( &remainder, &last_a );

	/* Compute a*denominator % b ... */
	compute_modulus( &conv_denom[num_computed_convergents],
			 &mod_value[num_computed_convergents] );

	++num_computed_convergents;
}


void
least_i_with_mod_le( goal, i, ai_mod_b )
	bigint		*goal, *i, *ai_mod_b;
{
	/* Returns the least i such that a*i % b <= goal.
	 * Also returns in ai_mod_b the value of a*i % b.
	 */
	
	bigint		distance_to_go, distance_we_went;
	bigint		decrement, quotient, remainder, kq1;
	int		last_conv_gt;
	int		new_conv;
	int		new_conv_compare;

	/* The odd convergents are those with a*q % b just a bit
	 * larger than 0.  These are the ones, then, that we want
	 * to deal with.
	 */

	/* Crank out convergents until we get one whose mod_value is <= goal. */

	new_conv = last_odd_conv - 2;
	do {
		/* Generate the next candidate convergent. */
		last_conv_gt = new_conv;
		new_conv += 2;

		while ( num_computed_convergents <= new_conv )
			next_convergent();

		new_conv_compare = bigint_compare( &mod_value[new_conv], goal );

	} while ( new_conv_compare > 0 );

	last_odd_conv = new_conv;

	if ( new_conv == 1 || new_conv_compare == 0 ) { /* All done. */
		bigint_copy( &conv_denom[new_conv], i );
		bigint_copy( &mod_value[new_conv], ai_mod_b );
		return;
	}

	/* The i we're looking for is somewhere between last_conv_gt
	 * and new_conv.  It will have the form q0 + k*q1, where q0 is
	 * the convergent denominator corresponding to last_conv_gt
	 * and q1 is the convergent denominator just after it.
	 */
	bigint_subtract( &mod_value[last_conv_gt], goal, &distance_to_go );

	/* Compute how much of a decrement in the mod_value we get with
	 * each addition of q1.  Note that a*q1 % b is very close to b,
	 * so that when we *add* q1 to some i, we *decrease* a*(i+q1) % b
	 * (unless a*i % b < b - (a*q1 % b), i.e., unless we're already
	 * closer to 0 than q1 can bring us ...).
	 */
	bigint_subtract( b, &mod_value[last_conv_gt + 1], &decrement );

	bigint_divide( &distance_to_go, &decrement, &quotient, &remainder );

	if ( ! bigint_is_zero( &remainder ) ) {
		/* Inexact divide: need k = quotient + 1. */
		bigint_add( &quotient, &big_one, &quotient );
	}

	/* Okay, now k = quotient.  Compute i and a*i % b and we're done. */
	bigint_multiply( &quotient, &conv_denom[last_conv_gt + 1], &kq1 );
	bigint_add( &conv_denom[last_conv_gt], &kq1, i );

	bigint_multiply( &decrement, &quotient, &distance_we_went );
	bigint_subtract( &mod_value[last_conv_gt], &distance_we_went,
			ai_mod_b );
}


void
least_i_with_mod_ge( goal, i, ai_mod_b )
	bigint		*goal, *i, *ai_mod_b;
{
	/* Returns the least i such that a*i % b >= goal.
	 * Also returns in ai_mod_b the value of a*i % b.
	 */
	
	bigint		distance_to_go, distance_we_went;
	bigint		*increment, quotient, remainder, kq1;
	int		last_conv_lt;
	int		new_conv;
	int		new_conv_compare;

	/* The even convergents are those with a*q % b just a bit
	 * smaller than b.  These are the ones, then, that we want
	 * to deal with.
	 */

	/* Crank out convergents until we get one whose mod_value is >= goal. */

	new_conv = last_even_conv - 2;
	do {
		/* Generate the next candidate convergent. */
		last_conv_lt = new_conv;
		new_conv += 2;

		while ( num_computed_convergents <= new_conv )
			next_convergent();

		if ( new_conv > 0 && bigint_is_zero( &mod_value[new_conv] ) )
			/* No further progress possible. */
			break;

		new_conv_compare = bigint_compare( &mod_value[new_conv], goal );

	} while ( new_conv_compare < 0 );

	last_even_conv = new_conv;

	if ( new_conv == 0 || new_conv_compare == 0 ) { /* All done. */
		bigint_copy( &conv_denom[new_conv], i );
		bigint_copy( &mod_value[new_conv], ai_mod_b );
		return;
	}

	/* The i we're looking for is somewhere between last_conv_lt
	 * and new_conv.  It will have the form q0 + k*q1, where q0 is
	 * the convergent denominator corresponding to last_conv_lt
	 * and q1 is the convergent denominator just after it.
	 */
	bigint_subtract( goal, &mod_value[last_conv_lt], &distance_to_go );

	/* Compute how much of a increment in the mod_value we get with
	 * each addition of q1.  Note that a*q1 % b is very close to 0,
	 * so that when we add q1 to some i, we increase a*(i+q1) % b
	 * (unless a*i % b > b - a*q1 % b, i.e., unless we're already
	 * closer to b than q1 can bring us ...).
	 */
	increment = &mod_value[last_conv_lt + 1];

	bigint_divide( &distance_to_go, increment, &quotient, &remainder );

	if ( ! bigint_is_zero( &remainder ) ) {
		/* Inexact divide: need k = quotient + 1. */
		bigint_add( &quotient, &big_one, &quotient );
	}

	/* Okay, now k = quotient.  Compute i and a*i % b and we're done. */
	bigint_multiply( &quotient, &conv_denom[last_conv_lt + 1], &kq1 );
	bigint_add( &conv_denom[last_conv_lt], &kq1, i );

	bigint_multiply( increment, &quotient, &distance_we_went );
	bigint_add( &mod_value[last_conv_lt], &distance_we_went, ai_mod_b );
}


void
closest_modulus_near_target( target, s, min_val, max_val, distance, near_zero,
				progress_fcn )
	bigint		*target, *s, *min_val, *max_val, *distance;
	int		near_zero;
	int		(*progress_fcn)();
{
	/* Finds an s with min_val <= s <= max_val such that
	 *	(a*s - target) % b
	 * is either minimal (as close to 0 as possible) or maximal
	 * (as close to b as possible), depending on whether near_zero
	 * is true or false, respectively.  Returns the final distance
	 * from the target in "distance".
	 *
	 * Each intermediate value from the starting point to the minimal/
	 * maximal s (including both the initial and final values of s)
	 * are reported in a call to progress_fcn, if non-null.  The
	 * function is given four arguments: the intermediate value of s,
	 * the modulus base b, the distance from the target, and a flag
	 * that is zero for each call save the one made when s is optimal
	 * (or no more progress can be made).  progress_fcn should return
	 * 0 if we are to continue trying to optimize s and 1 if we
	 * should give up.
	 *
	 * Assumes that init_for_convergents() has already been called.
	 */

	bigint		modulus, i, i_modulus;
	bigint		quotient, remainder, delta;

	bigint_copy( min_val, s );
	compute_modulus( s, &modulus );

	bigint_subtract( target, &modulus, distance );

	/* Since we're beginning a new minimization problem, reset the
	 * indications of how many convergents we have used so far.  Note
	 * that strictly speaking we need to only set one of these,
	 * depending on the value of near_zero, but it can't hurt to set
	 * them both.
	 */
	last_even_conv = 0;
	last_odd_conv = 1;

	while ( ! bigint_is_zero( distance ) ) {
		/* NOTE: in the comments the follow, the notation {expr1,expr2}
		 * means "if near_zero then expr1 else expr2".
		 */

		if ( progress_fcn && (*progress_fcn)( s, distance, b, 0 ) )
			return;

		if ( distance->bsign == negative )
			bigint_add( distance, b, distance );

		if ( near_zero )
			{
			least_i_with_mod_ge( distance, &i, &i_modulus );

			/* Convert i_modulus into a value that reflects
			 * how close each multiple of i takes us to our
			 * target.
			 */
			bigint_subtract( b, &i_modulus, &i_modulus );
			}
		else
			least_i_with_mod_le( distance, &i, &i_modulus );

		/* Okay, s' = s+i is the least value >= min_val such that
		 * {target <= a*s' % b < a*s % b, a*s' % b > a*s % b}.
		 *
		 * The case of a*s' % b = a*s % b will not arise since
		 * it implies that distance = 0.
		 *
		 * Furthermore, we know that
		 *
		 *	(a*s' % b) = (a*s % b) {-,+} modulus_i
		 *
		 * so we can keep adding multiples of i to s as long as
		 * a*(s+ki) % b {>,<} target.  Compute how many multiples
		 * we can fruitfully use, making sure we don't wind up
		 * with an s' > max_val.
		 */

		if ( near_zero ) {
			/* Change distance to how reflect how far away
			 * we are from the target, rather than
			 * (target - modulus) % b.
			 */
			bigint_subtract( b, distance, distance );
		}
		bigint_divide( distance, &i_modulus, &quotient, &remainder );

		/* We can add up to "quotient" number of i's to s ... */

		bigint_multiply( &i, &quotient, &delta );
		bigint_add( s, &delta, s );

		/* Compute the change in distance we just effected. */
		bigint_multiply( &i_modulus, &quotient, &delta );
		bigint_subtract( distance, &delta, distance );

		if ( bigint_compare( s, max_val ) > 0 ) {
			/* Oops - we went too far.  Remove multiples of i
			 * till we're back to being <= max_val.  At this
			 * point, we can't make any improvement, so we're
			 * done.
			 */

			bigint		distance_too_far;

			/* How much too far did we go? */
			bigint_subtract( s, max_val, &distance_too_far );

			/* How many multiples of i will bring us back in
			 * range?
			 */
			bigint_divide( &distance_too_far, &i,
					&quotient, &remainder );

			bigint_multiply( &i, &quotient, &delta );
			bigint_subtract( s, &delta, s );

			bigint_multiply( &i_modulus, &quotient, &delta );
			bigint_add( distance, &delta, distance );

			if ( ! bigint_is_zero( &remainder ) ) {
				/* Need to deduct one more multiple of i
				 * to get s <= max_val.
				 */
				bigint_subtract( s, &i, s );
				bigint_add( distance, &i_modulus, distance );
			}

			break;
		}

		if ( near_zero )
			/* Distance holds modulus - target; we want it
			 * to hold target - modulus
			 */
			distance->bsign = negative;
	}

	if ( progress_fcn )
		(void) (*progress_fcn)( s, distance, b, 1 );
}
