/* Testbase common routines for manipulating single and double precision */

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


/* Routines which should be the same for single and double precision
 * in most environments.
 */


#include "testbase.h"


void
binary_to_decimalstring(x, s, n, t)
	FLOAT          *x;
	decimalstring   s;
	unsigned long   n;
	enum testtype   t;

{
	/*
	 * Using standard C sprintf, converts binary to decimal string in
	 * format according to parameter n, which might be n significant
	 * digits for E format, n digits after point for F format, n
	 * significant digits for G format
	 */
	int		len = n;

	/* Following sample code tests sprintf.	 */

	switch (t) {
	case btoe:
	case etob:
		/* Standard C E format. */
		(void) sprintf(s, "%*.*e", len + 8, len - 1, (double) *x);
		break;
	case btof:
	case ftob:
		/* Standard C F format. */
		(void) sprintf(s, "%*.*f", len + 8, len, (double) *x);
		break;
	case btog:
	case gtob:
		/* Alternate C G format - need to force dot and trailing
		 * zeros in order to get ulps in the right place during
		 * mono checks, to avoid accidental overflow.
		 */
		(void) sprintf(s, "%#*.*g", len + 8, len, (double) *x);
		break;

	default:
		fatal("bad test type in binary_to_decimalstring()");
	}
}

void
decimalstring_to_binary(s, x)
	char           *s;
	FLOAT          *x;

{
	/*
	 * Using standard C sscanf, converts decimal string to binary.
	 */

	/* Following sample code tests sscanf.	 */

	(void) sscanf(s, CONV, x);
}

void
find_floatsize()
{
	/* Determine size of FLOAT. */

	floatsizeunion  u;
	int             i;

	floatsize = 0;
	for (i = 0; i < MAXFLOATSIZE; i++)
		u.fsu[i] = 0;
	for (i = 0; i < 10; i++) {
		u.f = urandom() / (double) urandom();
		while (floatsize < MAXFLOATSIZE && u.fsu[floatsize] != 0)
			floatsize++;
	}
	if ((floatsize * sizeof(FLOATSIZEUNIT)) != sizeof(FLOAT)) {
		fatal(
	"dynamic binary size does not agree with sizeof in find_floatsize()");
	}
}

void
find_signoise()
{
	/*
	 * Determine smallest n such that 1 is negligible with respect to
	 * 10**n.
	 */

	FLOAT           one;
	FLOAT           big, bigplus, bigminus;

	big = 1;
	one = 1;
	signoise = 0;
	do {
		big = big * 10;
		signoise++;
		bigplus = big + one;
		bigminus = big - one;
		dummy(&bigplus);
		dummy(&bigminus);
	} while ((bigplus != big) && (bigminus != big));
}

void
find_lsuindex()
{
	/* Determine index of FLOATSIZEUNIT that points to lsu of FLOAT. */

	floatsizeunion  u1, u1plus;
	FLOAT           x, d;
	int             i;

	x = 1;
	u1.f = x;
	d = 1;
	do {
		u1plus.f = x;
		d = d / 2;
		x = 1 + d;
		dummy(&x);
	} while (x != 1);
	i = 0;
	while (u1.fsu[i] == u1plus.fsu[i])
		i++;
	if (i < floatsize)
		lsuindex = i;
	else
		fatal("couldn't find lsu; no units changed in find_lsuindex()");

	i++;
	while (u1.fsu[i] == u1plus.fsu[i])
		i++;
	if (i < floatsize)
		fatal(
		"couldn't determine lsu; > 1 unit changed in find_lsuindex()");

	{
		union {
			FLOAT           f;
			struct {
				unsigned        s:1;
				unsigned:       31;
			}               fsu;
		}               u;
		u.f = x;
		signbitfirst = u.fsu.s;
		u.f = -x;
		if (signbitfirst == u.fsu.s)
			signbitfirst = 0;
		else
			signbitfirst = 1;
	}
}

void
find_f_exponents()
{
	min_10_f_exponent = -(int) cursig;
	min_2_f_exponent = -(int) cursig * LOG_10_TO_BASE_2;
	max_10_f_exponent = DECIMALSTRING_LENGTH - cursig - 3;
	max_2_f_exponent = max_10_f_exponent * LOG_10_TO_BASE_2;
}

/* Some systems generate fatal faults on overflow, in which case
 * this routine won't work.
 */
#ifndef NO_FIND_EXTREMES
void
find_extremes()
{
	/* Find extreme exponents.		 */

	FLOAT           x, y;
	EXPONENT        ix;

	x = 1;
	y = 10;
	ix = 0;
	do {
		ix++;
		x = y;
		y = 10 * x;
		dummy(&y);
	} while (ABS(y / x - 10) < 1e-3);	/* 1e-3 is pretty arbitrary. */

	/*
	 * Back off by one to avoid getting too close to overflow threshold
	 * when adding ulps.
	 */
	max_ten_exponent = ix - 1;

	x = 1;
	y = x / 10;
	ix = 0;
	do {
		ix++;
		x = y;
		y = x / 10;
		dummy(&y);
	} while (ABS(y / x - 0.1) < 1e-3);	/* 1e-3 is pretty arbitrary. */
	min_ten_exponent = -ix;

	x = 1;
	y = 2;
	ix = 0;
	do {
		ix++;
		x = y;
		y = 2 * x;
		dummy(&y);
	} while (ABS(y / x - 2) < 1e-3);	/* 1e-3 is pretty arbitrary. */

	/*
	 * Back off by one to avoid getting too close to overflow threshold
	 * when adding ulps.
	 */
	max_two_exponent = ix - 1;

	x = 1;
	y = x / 2;
	ix = 0;
	do {
		ix++;
		x = y;
		y = x / 2;
		dummy(&y);
	} while (ABS(y / x - 0.5) < 1e-3);	/* 1e-3 is pretty arbitrary. */
	min_two_exponent = -ix;
}
#endif	/* ! NO_FIND_EXTREMES */

void
copyfloat(x, y)
	FLOAT          *x, *y;

{
	/* Used to copy FLOAT variables: *y = *x.	 */

	*y = *x;
}

#ifndef NO_CONVERSION_ROUTINES
void
testbd(x, s, n, t, e)
	FLOAT          *x;
	decimalstring   s;
	unsigned long  *n;
	enum testtype  *t;
	int            *e;

{
	/*
	 * Using routine to be tested, converts binary to decimal string in
	 * format according to parameter n, which might be n significant
	 * digits for E format, n digits after point for F format, n
	 * significant digits for G format
	 */
	int		len = *n;

	clear_exceptions();

	/* Following sample code tests sprintf.	 */

	switch (*t) {
	case btoe:
		/* Standard C E format. */
		(void) sprintf(s, "%*.*e", len + 8, len - 1, (double) *x);
		break;

	case btof:
		/* Standard C F format. */
		(void) sprintf(s, "%*.*f", len + 8, len, (double) *x);
		break;

	case btog:
		/* Standard C G format. */
		(void) sprintf(s, "%*.*g", len + 8, len, (double) *x);
		break;

	default:
		fatal("bad test type in testbd()");
	}

	*e = get_exceptions();
}

void
testdb(s, x, e)
	char           *s;
	FLOAT          *x;
	int            *e;

{
	/*
	 * Using routine to be tested, converts decimal string to binary.
	 */

	clear_exceptions();
	(void) sscanf(s, CONV, x);
	*e = get_exceptions();
}
#endif	/* ! NO_CONVERSION_ROUTINES */
