/* Testbase routines for generating stress tests */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */


#include "testbase.h"


/* The minimum and maximum values that a normalized significand in base b1
 * can have.  These values are b1^(d1-1) and b1^d1 - 1, respectively.
 */
static bigint	b1_min_normal, b1_max_normal;

static EXPONENT	e1;	/* exponent of conversion source */

/* We remember the distance and modulus associated with the current
 * stress test in these globals.  Then excess_bits_in_stress_test()
 * can use these values to compute the number of excess bits required
 * by the test.
 */
static bigint	*current_distance, *current_modulus;

/* When we're trying to approach a value whose fraction is near a half-ulp,
 * we do so by setting the target to be one-half of the full-ulp value.
 * If the full-ulp value is odd then the target will be actually just a
 * little off (below, since we use integer division and round down) from
 * exactly one-half.  This flag keeps track of whether the target is such
 * a rounded-down half-ulp.  If so, then if we exactly attain the target
 * excess_bits_in_stress_test() needs to be aware that the value does not
 * indicate 0 stress bits (since the target was exactly obtained) but
 * rather a maximal number of stress bits, since what we actually obtained
 * is just slightly off from an exact half-ulp.
 */
static int	target_is_inexact;

void
compute_min_and_max_normals( base, d, min_normal, max_normal )
	unsigned	base;
	int		d;
	bigint		*min_normal, *max_normal;
{
	if ( base == 2 ) {
		two_to_power( (EXPONENT) (d - 1), min_normal );
		two_to_power( (EXPONENT) d, max_normal );
	} else {
		bigint_copy( ten_to_power( (EXPONENT) (d - 1) ), min_normal );
		bigint_copy( ten_to_power( (EXPONENT) d ), max_normal );
	}

	bigint_subtract( max_normal, &big_one, max_normal );
}


void
compute_power_fraction( b1, e1, b2, e2, numerator, denominator )
	unsigned	b1, b2;
	EXPONENT	e1, e2;
	bigint		*numerator, *denominator;
{
	/* Returns a fraction with integer numerator and denominator equal
	 * to b1^e1*b2^e2, where e1 or e2 might be negative.
	 *
	 * Note that upon return, gcd(numerator,denominator) = 1.
	 */

	int		num_numerator_powers, num_denominator_powers;
	EXPONENT	removeable_powers_of_two;

	bigint_set( numerator, 1 );
	bigint_set( denominator, 1 );

	num_numerator_powers = 0;
	num_denominator_powers = 0;

	if ( e1 >= 0 ) {
		/* Scale up numerator.. */
		bigint_times_power( numerator, b1, e1 );
		num_numerator_powers += e1;
	} else {
		bigint_times_power( denominator, b1, -e1 );
		num_denominator_powers += -e1;
	}

	if ( e2 >= 0 ) {
		bigint_times_power( numerator, b2, e2 );
		num_numerator_powers += e2;
	} else {
		bigint_times_power( denominator, b2, -e2 );
		num_denominator_powers += -e2;
	}

	/* Each power, whether the base is 2 or 10, contributes a
	 * power of two to the numerator or denominator that it's
	 * multiplied into.  Compute how many such powers of two
	 * we can remove from both numerator and denominator.
	 */
	removeable_powers_of_two = 
		num_numerator_powers < num_denominator_powers ?
			num_numerator_powers : num_denominator_powers;

	bigint_times_2_to_power( numerator, -removeable_powers_of_two );
	bigint_times_2_to_power( denominator, -removeable_powers_of_two );
}


void
find_b1_range( b1, b2, d1, e2, e1min_addr, e1max_addr, imin, imax )
	unsigned	b1, b2;
	int		d1;
	EXPONENT	e2, *e1min_addr, *e1max_addr;
	bigint		*imin, *imax;
{
	EXPONENT	e1min, e1max;
	WORK		log_b2_over_log_b1, temp;
	bigint		numerator, denominator, remainder;

	log_b2_over_log_b1 = LOG_10_TO_BASE_2;
	if ( b2 == 2 )
		log_b2_over_log_b1 = 1.0 / log_b2_over_log_b1;


	temp = (WORK) e2 * log_b2_over_log_b1 - 1.0;

	/* Compute e1min = ceil(temp). */
	e1min = temp;
	if ( e1min < temp )
		++e1min;

	temp = ((WORK) e2 + 1.0) * log_b2_over_log_b1;

	/* Compute e1max = floor(temp). */
	e1max = temp;
	if ( e1max > temp )
		--e1max;


	/* Now compute imin and see whether it's in the proper range.  If
	 * not, we may need to adjust e1min up or down and try again.  imin
	 */
	for (;;) {
		/* Compute b1^(d1-1) * b2^e2 / b1^e1min.
		 * This value is equivalent to b1^(d1-1-e1min) * b2^e2.
		 */

		compute_power_fraction( b1, d1 - 1 - e1min, b2, e2,
			&numerator, &denominator );

		bigint_divide( &numerator, &denominator, imin, &remainder );

		if ( ! bigint_is_zero( &remainder ) ) {
			/* Need to add one to imin to effect ceil() */
			bigint_add( imin, &big_one, imin );
		}

		/* Check that imin is in the proper range for normalization.
		 * If not then initial guess for e1min was wrong (this should
		 * only happen for the case of e2 = 0, or if d1 is very small,
		 * though it could also happen due to faulty rounding in
		 * computing log_b2_over_log_b1).  Adjust it and repeat.
		 */

		if ( bigint_compare( imin, &b1_min_normal ) < 0 )
			/* imin is too small => e1min is too large */
			--e1min;

		else if ( bigint_compare( imin, &b1_max_normal ) > 0 )
			/* imin is too large => e1min is too small */
			++e1min;

		else
			/* imin is correct. */
			break;
	}

	/* Do the same for imax. */
	for (;;) {
		/* Compute b1^(d1-1) * (b2^(e2+1)) / b1^e1max.
		 * This value is equivalent to b1^(d1-1-e1max) * b2^(e2+1).
		 */

		compute_power_fraction( b1, d1 - 1 - e1max, b2, e2 + 1,
			&numerator, &denominator );

		bigint_divide( &numerator, &denominator, imax, &remainder );

		if ( bigint_is_zero( &remainder ) ) {
			/* Oops - we have imax*b1^(d1-1-e1max) = b2^(e2+1).
			 * But it has to be at least one less, in order
			 * to remain normalized.  See the paper!
			 */
			bigint_subtract( imax, &big_one, imax );
		}

		/* Check that imax is in the proper range for normalization.
		 * If not then initial guess for e1max was wrong this should
		 * really only happen for the case of e2 = -1, though it
		 * could also happen due to faulty rounding in computing
		 * log_b2_over_log_b1.  Adjust it and repeat.
		 */

		if ( bigint_compare( imax, &b1_min_normal ) < 0 )
			/* imax is too small => e1max is too large */
			--e1max;

		else if ( bigint_compare( imax, &b1_max_normal ) > 0 )
			/* imax is too large => e1max is too small */
			++e1max;

		else
			/* imax is correct. */
			break;
	}

	*e1min_addr = e1min;
	*e1max_addr = e1max;
}


void
worst_cases( b1, b2, d1, d2, e2, half_ulp )
	unsigned	b1, b2;
	int		d1, d2;
	EXPONENT	e2;
	int		half_ulp;
{
	/* Find worst cases (closest to an ulp or half-an-ulp [depending
	 * on whether half_ulp is set] but not exactly equal) of converting
	 * d1 digits, base b1 to d2 digits, base b2, where the exponent of
	 * the b2 form is = e2.
	 *
	 * b1 and b2 must be 2 or 10.
	 */

	EXPONENT	e1min, e1max;
	bigint		imin, imax, p, q, quotient, target;
	bigint		*min_val, *max_val;
	bigint		just_under, distance_under, just_over, distance_over;

	compute_min_and_max_normals( b1, d1, &b1_min_normal, &b1_max_normal );

	find_b1_range( b1, b2, d1, e2, &e1min, &e1max, &imin, &imax );

	for ( e1 = e1min; e1 <= e1max; ++e1 ) {
		compute_power_fraction( b1, e1-d1+1, b2, d2-1-e2, &p, &q );

		if ( bigint_compare( &p, &q ) >= 0 ) {
			/* p > q: set p = p % q */
			bigint_divide( &p, &q, &quotient, &p );
		}

		min_val = (e1 == e1min ? &imin : &b1_min_normal);
		max_val = (e1 == e1max ? &imax : &b1_max_normal);

		if ( bigint_is_zero( &p ) ) {
			/* All values are exactly representable.  Pick
			 * the end cases for testing.
			 */
			(void) next_toughie( min_val, &big_zero, &q, 1 );
			(void) next_toughie( max_val, &big_zero, &q, 1 );
			continue;
		}

		target_is_inexact = 0;

		if ( half_ulp ) {
			/* Our target is half an ulp, i.e., q/2 */
			bigint_copy( &q, &target );
			bigint_times_2_to_power( &target, (EXPONENT) -1 );

			if ( (q.bsignificand[0] & 0x1) != 0 )
				/* q is odd, so q/2 is inexact. */
				target_is_inexact = 1;

		} else
			/* Target is an ulp, i.e., 0. */
			bigint_set( &target, 0 );

		init_for_convergents( &p, &q );

		closest_modulus_near_target( &target, &just_under,
			min_val, max_val, &distance_under, 0, next_toughie );
		closest_modulus_near_target( &target, &just_over,
			min_val, max_val, &distance_over, 1, next_toughie );

	}
}

int
next_toughie( s, distance, b, is_best )
	bigint		*s, *distance, *b;
	int		is_best;
{
	if ( is_to_binary( curtest ) ) {
		decimalstring d;

		bigint_to_decimalstring( s, d );
		sprintf( center.d, "%c%c.%se%ld",
			 random_sign(), d[0], &d[1], e1 );

	} else {
		unpacked u;

		u.usign = (random_sign() == '+') ? positive : negative;
		u.uclass = number;
		u.ubase = 2;
		u.uexponent = e1 - FLOAT_PRECISION + 1;
		bigint_copy( s, &u.usignificand );

		unpacked_to_binary( &u, FLOAT_ADDR center.b );
	}

	center.dt = stress;
	center.index = countregions;
	center.offset = 0;

	current_distance = distance;
	current_modulus = b;

	(void) one_test( &center );

	return 0;
}

EXPONENT
excess_bits_in_stress_test()
{
	/* Return the number of excess bits required by the current
	 * stress test.
	 */
	bigint		quotient, remainder;

	if ( bigint_is_zero( current_distance ) ) {
		if ( target_is_inexact )
			/* current_distance is really .5 ... */
			return bigint_num_bits( current_modulus ) + 1;

		return 0;
	}

	bigint_divide( current_modulus, current_distance,
			&quotient, &remainder );

	return bigint_num_bits( &quotient );
}
