/* Testbase main program */

#ifndef lint
static char     sccsid[] = "@(#)testbase.c 1.14 91/02/28 SMI";
#endif

/*
 * Copyright (c) 1989 by Sun Microsystems, Inc.
 */

/*********************************************************************

This program is for testing conversion between internal binary
floating-point format and ASCII decimal floating-point strings.
Notice the copyright above; persons receiving copies of the program
for evaluation are not to distribute it.

Evaluators are encouraged to investigate the correctness of this program
and its applicability to IEEE and other reasonable computer systems.
Suggestions for tests related to those already implemented and for user
interface are also of interest.  Efficiency and source code formatting
issues are not presently of interest.

Testbase has the following options:
	-all	all tests implemented
	-btoe	binary to E format
	-btof	binary to F format
	-btog	binary to G format
	-etob	E format to binary
	-ftob	F format to binary
	-gtob	G format to binary

Rounding modes to test; default none:
	-ieee		all four IEEE rounding modes
	-round nearest	IEEE unbiased round to nearest
	-round zero	IEEE round toware zero
	-round up	IEEE round toward positive infinity
	-round down	IEEE round toward negative infinity
	-round biased	biased rounding to nearest such as VAX
	-round out	rounding toward infinity, away from zero
	-round one	rounding that guarantees < 1 ulp
	-round two	rounding that guarantees < 2 ulp

Test parameters:
	-sig n		significance parameter, default decimal precision+1
	-sigmin n	minimum significance parameter, default 1
	-sigmax n	maximum significance parameter, default 2*precision
	-regions n	number of regions to check monotonicity, default
			varies with exponent range
	-time n		number of secs. to spend on each test, default no limit
	-monotests n	number of points on either side of center in each
			region, default 2
	-terse		minimal output, default is terse
	-verbose	more output
	-tedious	drown in output
	-random		generate only random data, default systematically
			generates
			powers of ten and two as well as random data
	-nozero		don't test sign of zero; for Fortran
	-noexceptions	don't test exceptions
	-nostress	don't do stress-testing
	
The program generates data, converts it according to a format and
significance specification, and compares the computed result to the
operand, computing the difference in ULPs of the computed result.  The
largest positive and negative ulp errors are noted, along with wrong-sign
errors and wrong-class errors.

By default all powers of bases 10 and 2 are generated as well as random
data and stress-testing.  Random data alternates among really random data
and data chosen to hit random halfway cases of the target.

The critical calculation of scaled ulps of error is exact, and done
entirely with "bigint" integer arithmetic.  The bigint scaled ulps
value is then converted to actual ulps of the target.

Since any representable floating-point number in base bt can be written

	it * bt ** et

for integers it and et, the error between two attempts vs and vt to
represent the same exact value in two different floating-point systems is
just

	error = it * bt ** et - is * bs ** es

and since a unit in the last place of vt is just (sign of it) * 1 * bt ** et,

	error in ulps of vt = (sign of it) * { it - is * bs ** es * bt ** -et }

Since es or -et could be negative, the right side might not be calculatable
in integer arithmetic.  But a scaled error in ulps can be computed
in integer arithmetic.  For instance, for large enough data, es and et are
both positive.  Therefore

	scaled ulps = ulps * bt ** et
		    = (sign of it) * { it * bt ** et - is * bs ** es }

and the right hand difference can be computed exactly using just integer
arithmetic, of possibly high precision.  Then the actual ulps could be
computed

	ulps = (scaled ulps) * bt ** -et

using normal floating-point arithmetic, but roundoff error could distort
critical cases near half an ulp, for instance.    Therefore actual ulp
differences are computed as a sum of exact half-ulps plus or minus a
residual of magnitude no greater than 0.25 ulp.  The residual is computed
subject to rounding error in WORKing precision.  Ulp differences are also
printed in the same format: n half-ulps plus a residual.

**********************************************************************/

#include "testbase.h"
#include <math.h>	/* sqrt() needed by find_good_stride() */


/* Globals - see testbase.h for descriptions. */

int		floatsize, lsuindex;
int		signbitfirst;
enum verbositytype	verbosity;
int		sigmin, sigmax, signoise;
int		no_exceptions, no_stress, no_sign_zero;
int		skip_powers;
int		introduce_errors, introduced_error;
enum roundingtype	curround, normal_rounding;
enum testtype	curtest;
int		test_enabled[(int) (lasttesttype)];
int		rounding_enabled[(int) (lastroundingtype)];
WORK		timestart, timestop, timelimit;
int		timedout;
enum datatype	lastdata;
EXPONENT	max_ten_exponent, min_ten_exponent;
EXPONENT	max_two_exponent, min_two_exponent;
EXPONENT	max_2_f_exponent, min_2_f_exponent;
EXPONENT	max_10_f_exponent, min_10_f_exponent;
EXPONENT	max_dec_stress_exponent, min_dec_stress_exponent;
EXPONENT	max_bin_stress_exponent, min_bin_stress_exponent;
EXPONENT	max_btof_dec_stress_exponent, min_btof_dec_stress_exponent;
EXPONENT	dec_stress_exponent_stride, bin_stress_exponent_stride;
EXPONENT	btof_dec_stress_exponent_stride;
EXPONENT        unpacked_one_exponent;
unsigned long	cursig;
unsigned long	nregions;
int		nmono;
WORK		ulpsbound;
testrecord	center, mo[2];
unsigned long	countregions, countclass, countexcep, countsign, countmono,
		countround, countconversions, countintro,
		count_intro_caught, count_intro_missed;
unsigned long	total_conversions, total_regions;
testrecord	excepexample, classexample, signexample;
testrecord	monoexample1, monoexample2;
testrecord	roundexample, whopperexample;
testrecord      ulpminpos, ulpmaxpos, ulpminneg, ulpmaxneg;
bigint		big_zero, big_one;


/*****************************************************************************

	GLOBAL CONSTANT STRUCTURES

******************************************************************************/

char	*roundingmodenames[lastroundingtype] =
{
	"unbiased",		/* round_unbiased,	IEEE default */
	"toward zero",		/* round_in,		IEEE to zero */
	"toward +infinity",	/* round_up,		IEEE to +infinity */
	"toward -infinity",	/* round_down,		IEEE to -infinity */
	"biased",		/* round_biased,	VAX */
	"toward infinity",	/* round_out,		IEEE to infinity */
	"error < 1 ulp",	/* round_one,		Error < 1 ulp. */
	"error < 2 ulp"		/* round_two,		Error < 2 ulp.  */
};

char	*testnames[lasttesttype] =
{
	"btoe",
	"btof",
	"btog",
	"etob",
	"ftob",
	"gtob",
};


void
dummy(pf)
	FLOAT          *pf;
{
	/*
	 * Routine to force storage of variables allocated to extended
	 * registers.
	 */
char * pu;

pu = (char *) pf;
*pu += 128;
*pu ^= 0x80;		/* Will this odd nop fool an inlining optimizer? */
}

void
workdummy(pf)
	WORK          *pf;
{
	/*
	 * Routine to force storage of variables allocated to extended
	 * registers.
	 */
char * pu;

pu = (char *) pf;
*pu += 128;
*pu ^= 0x80;		/* Will this odd nop fool an inlining optimizer? */
}

EXPONENT
gcd( x, y )
	EXPONENT	x, y;
{
	EXPONENT	t;

	while ( y > 0 ) {
		t = x % y;
		x = y;
		y = t;
	}

	return x;
}

EXPONENT
find_good_stride( range )
	EXPONENT	range;
{
	/* Returns a good "stride" for stepping through the [0,range).
	 * To be "good" the stride should be near the square-root of
	 * range (so we get to the different regions of range fairly
	 * quickly) and have gcd(stride,range) = 1, so that
	 *
	 *	i*stride % range	1 <= i <= range
	 *
	 * will visit each value in [0,range) exactly once.
	 */
	EXPONENT	stride, offset, increment;

	if ( range == 0 )
		fatal( "empty range in find_good_stride()" );

	stride = (EXPONENT) sqrt( (double) range );

	/* If the range is even then make sure the stride isn't. */
	if ( range % 2 == 0 ) {
		if ( stride % 2 == 0 )
			++stride;
		increment = 2;	/* subsequently, only try odd strides */
	} else
		increment = 1;

	/* Loop till gcd(stride,range) = 1. */
	for (offset = 0; ; offset += increment) {
		if ( gcd( range, stride + offset ) == 1 ) {
			stride += offset;
			break;
		}
		if ( gcd( range, stride - offset ) == 1 ) {
			stride -= offset;
			break;
		}
	}

	return stride;
}

void
interrupt()
{	/* Called when an interrupt is received. */
	summary();
	exit(0);
}

void
fatal(s)
	char           *s;
{
	(void) fprintf(stderr, "fatal internal error: %s\n", s);
	exit(1);
}

void usage()
{
	fprintf( stderr, "testbase [-all -btoe -btof -btog -etob -ftob -gtob\n" );
	fprintf( stderr, "          -ieee -nozero -noexceptions -random\n" );
	fprintf( stderr, "          -tedious -terse -verbose]\n" );
	fprintf( stderr, "         [-monotests #] [-regions #]\n" );
	fprintf( stderr,
	"         [-sig #] [-sigmin #] [-sigmax #] [-time seconds]\n" );
	fprintf( stderr,
	    "         [-round {nearest up down zero biased out one two}]\n" );
	exit( 1 );
}


void bad_flag( flag )
	char		flag[];
{
	fprintf( stderr, "bad flag - \"%s\"\n", flag );
	usage();
}


void bad_flag_arg( flag )
	char		flag[];
{
	fprintf( stderr, "flag \"%s\" requires an argument\n", flag );
	usage();
}

void
parse_args(argc, argv)
	int             argc;
	char           *argv[];

{
	int             i, j;

	for (i = 1; i < argc; i++) {
		switch (argv[i][1]) {

		case 'a':
			/* -all to run all tests */
			if (strcmp(argv[i], "-all") == 0) {
				for (j = 0; j < (int) lasttesttype; j++)
					test_enabled[j] = 1;

			} else
				bad_flag( argv[i] );
			break;

		case 'b':
			/* -bto[efg] to run that test */
			if (strcmp(argv[i], "-btoe") == 0)
				test_enabled[(int) btoe] = 1;

			else if (strcmp(argv[i], "-btof") == 0)
				test_enabled[(int) btof] = 1;

			else if (strcmp(argv[i], "-btog") == 0)
				test_enabled[(int) btog] = 1;

			else
				bad_flag( argv[i] );
			break;

		case 'e':
			/* -etob to run that test */
			if (strcmp(argv[i], "-etob") == 0)
				test_enabled[(int) etob] = 1;

			else if (strcmp(argv[i], "-errors") == 0) {
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );
				(void) sscanf(argv[i], "%d", &introduce_errors);

			} else
				bad_flag( argv[i] );
			break;

		case 'f':
			/* -ftob to run that test */
			if (strcmp(argv[i], "-ftob") == 0)
				test_enabled[(int) ftob] = 1;
			else
				bad_flag( argv[i] );
			break;

		case 'g':
			/* -gtob to run that test */
			if (strcmp(argv[i], "-gtob") == 0)
				test_enabled[(int) gtob] = 1;

			else
				bad_flag( argv[i] );
			break;

		case 'i':
			/* -ieee to enable IEEE rounding modes */
			if (strcmp(argv[i], "-ieee") == 0) {
				rounding_enabled[(int) round_unbiased] = 1;
				rounding_enabled[(int) round_up] = 1;
				rounding_enabled[(int) round_down] = 1;
				rounding_enabled[(int) round_in] = 1;

			} else
				bad_flag( argv[i] );
			break;

		case 'm':
			if (strcmp(argv[i], "-monotests") == 0) {
				/*
				 * -monotests n to make n the number of
				 * adjacent representable points on either
				 * side of the center to be tested in each
				 * region
				 */
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );
				(void) sscanf(argv[i], "%d", &nmono);

			} else
				bad_flag( argv[i] );
			break;

		case 'n':
			if (strcmp(argv[i], "-nozero") == 0) {
				/*
				 * -nozero to prevent testing sign of zero
				 * for Fortran
				 */
				no_sign_zero = 1;

			} else if (strcmp(argv[i], "-nostress") == 0) {
				/*
				 * -nostress to prevent stress-testing
				 */
				no_stress = 1;
				nregions = -32 * min_ten_exponent;

			} else if (strcmp(argv[i], "-noexceptions") == 0) {
				/*
				 * -noexceptions to prevent testing IEEE
				 * arithmetic exceptions
				 */
				no_exceptions = 1;

			} else
				bad_flag( argv[i] );
			break;

		case 'r':
			if (strcmp(argv[i], "-random") == 0) {
				/*
				 * -random to bypass testing powers of ten
				 * and two, for debugging
				 */
				skip_powers = 1;

			} else if (strcmp(argv[i], "-regions") == 0) {
				/*
				 * -regions n to make n the number of regions
				 * to be tested at each significance
				 */
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );
				(void) sscanf(argv[i], "%lu", &nregions);

			} else if (strcmp(argv[i], "-round") == 0) {
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );

				else if (strcmp(argv[i], "nearest") == 0)
					rounding_enabled[(int) round_unbiased] =
						1;

				else if (strcmp(argv[i], "up") == 0)
					rounding_enabled[(int) round_up] = 1;

				else if (strcmp(argv[i], "down") == 0)
					rounding_enabled[(int) round_down] = 1;

				else if (strcmp(argv[i], "zero") == 0)
					rounding_enabled[(int) round_in] = 1;

				else if (strcmp(argv[i], "biased") == 0)
					rounding_enabled[(int) round_biased] = 1;
				else if (strcmp(argv[i], "out") == 0)
					rounding_enabled[(int) round_out] = 1;

				else if (strcmp(argv[i], "one") == 0)
					rounding_enabled[(int) round_one] = 1;

				else if (strcmp(argv[i], "two") == 0)
					rounding_enabled[(int) round_two] = 1;

				else
					bad_flag( argv[i] );

			} else
				bad_flag( argv[i] );
			break;

		case 's':
			if (strcmp(argv[i], "-sig") == 0) {
				/*
				 * -sig n to set min and max significance test
				 */
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );

				(void) sscanf(argv[i], "%d", &sigmin);
				sigmax = sigmin;

			} else if (strcmp(argv[i], "-sigmin") == 0) {
				/*
				 * -sigmin n to set maximum significance test
				 * to n
				 */
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );
				(void) sscanf(argv[i], "%d", &sigmin);

			} else if (strcmp(argv[i], "-sigmax") == 0) {
				/*
				 * -sigmax n to set maximum significance test
				 * to n
				 */
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );
				(void) sscanf(argv[i], "%d", &sigmax);

			} else
				bad_flag( argv[i] );
			break;

		case 't':
			if (strcmp(argv[i], "-tedious") == 0) {
				/*
				 * -tedious for complete listing of intermediate
				 * results, mainly for debugging
				 */
				verbosity = tedious;

			} else if (strcmp(argv[i], "-terse") == 0)
				/* -terse for listing of summaries only */
				verbosity = terse;

			else if (strcmp(argv[i], "-time") == 0) {
				/*
				 * -time n to set maximum time per test to n
				 * seconds
				 */
				long unsigned   itime;
				if ( ++i >= argc )
					bad_flag_arg( argv[i-1] );

				(void) sscanf(argv[i], "%lu", &itime);
				timelimit = itime;

			} else
				bad_flag( argv[i] );
			break;
		case 'v':
			if (strcmp(argv[i], "-verbose") == 0)
				/*
				 * -verbose for partial listing of intermediate
				 * results
				 */
				verbosity = verbose;

			else
				bad_flag( argv[i] );
			break;
		}
	}
}

void
init( argc, argv )
	int		argc;
	char		*argv[];
{
	int             i;
	decimalstring   done;
	FLOAT           bone;
	unpacked        uone;

	init_interrupt_handling();

	bigint_set( &big_zero, 0 );
	bigint_set( &big_one, 1 );

	/* Dynamically determine parameters of system where possible.	 */
	find_floatsize();
	find_signoise();
	find_lsuindex();
	find_extremes();

	tenpowerstring((EXPONENT) 0, (EXPONENT) 1, done);
	decimalstring_to_binary(done, FLOAT_ADDR bone);
	binary_to_unpacked(FLOAT_ADDR bone, &uone);
	unpacked_one_exponent = uone.uexponent;

	/* Set up default values. */

	verbosity = terse;

	max_bin_stress_exponent = max_two_exponent;
	min_bin_stress_exponent = min_two_exponent - unpacked_one_exponent;

	bin_stress_exponent_stride =
		find_good_stride(
			max_bin_stress_exponent - min_bin_stress_exponent + 1 );

	max_dec_stress_exponent = max_ten_exponent;

	/* The following is a somewhat crude attempt to guarantee that we
	 * don't try testing exponents that are so small they will result
	 * in subnormal results.  What we really ought to do is convert
	 * unpacked_one_exponent into decimal digits. ###
	 */
	min_dec_stress_exponent =
		MAX(min_ten_exponent, -max_dec_stress_exponent);

	dec_stress_exponent_stride =
		find_good_stride(
			max_dec_stress_exponent - min_dec_stress_exponent + 1 );

	nregions = 0;

	/* Test regions for each significance parameter. */
	nmono = 2;		/* Number of ulps on each side of center per
				 * monotonicity region. */

	sigmin = sigmax = signoise + 1;

	skip_powers = 0;
	timelimit = 0.0;	/* No time limit by default. */
	no_sign_zero = 0;	/* Test sign of zero by default. */
	no_exceptions = ! TEST_EXCEPTIONS_DEFAULT;
	no_stress = 0;		/* Do stress-testing by default. */

	introduce_errors = 0;	/* Don't introduce errors by default. */

	normal_rounding = NORMAL_ROUNDING;

	for (i = 0; i < (int) lasttesttype; i++)
		test_enabled[i] = 0;	/* Default no tests. */

	for (i = 0; i < (int) lastroundingtype; i++)
		rounding_enabled[i] = 0;	/* Default no roundings. */


	parse_args( argc, argv );


	for (i = 0; i < (int) lasttesttype; i++)
		if ( test_enabled[i] )
			break;

	if ( i >= (int) lasttesttype ) {
		/* No test type(s) specified; default to etob, btoe */
		test_enabled[(int) etob] = 1;
		test_enabled[(int) btoe] = 1;
	}

	for (i = 0; i < (int) lastroundingtype; i++)
		if ( rounding_enabled[i] )
			break;

	if ( i >= (int) lastroundingtype )
		/* No rounding type(s) specified; default to normal. */
		rounding_enabled[(int) NORMAL_ROUNDING] = 1;


	(void) printf("Begin base conversion test version %s\n",
		"testbase.c 2.0 91/05/01");
	(void) printf("Copyright (c) 1988 by Sun Microsystems, Inc.\n");
	(void) printf("Type %s size is %d units of size %u.\n",
		FLOATSTRING, floatsize, sizeof(FLOATSIZEUNIT));
	(void) printf(
		"Type %s appears to be less significant than %d decimals.\n",
			FLOATSTRING, signoise);

	if (lsuindex == (floatsize - 1))
		(void) printf("Most significant unit stored first; ");
	else
		(void) printf("Least significant unit stored first; ");

	if (signbitfirst == 1)
		(void) printf("sign bit first in unit.\n");
	else
		(void) printf("sign bit last in unit.\n");

	(void) printf("Binary powers range from 2**%+ld to 2**%+ld\n",
		min_two_exponent, max_two_exponent);
	(void) printf("Decimal powers range from 10**%+ld to 10**%+ld\n",
		min_ten_exponent, max_ten_exponent);

	(void) printf("The unpacked exponent of 1.0 is %ld.\n",
		unpacked_one_exponent);


	(void) printf("Conversions to be tested:");
	for (i = 0; i < (int) lasttesttype; i++)
		if ( test_enabled[i] )
			(void) printf(" %s", testnames[i]);
	(void) printf("\n");

	(void) printf("Rounding modes to be tested:");
	for (i = 0; i < (int) lastroundingtype; i++)
		if ( rounding_enabled[i] )
			(void) printf(" %s", roundingmodenames[i]);
	(void) printf("\n");

	if ( no_stress || nregions > 0 )
		(void) printf("Test %lu regions.\n", nregions);
	else {
		if ( test_enabled[(int) btoe] || test_enabled[(int) btog] )
			(void) printf("Test %ld regions binary -> decimal.\n",
				max_dec_stress_exponent -
				min_dec_stress_exponent + 1);

		if ( test_enabled[(int) btof] ) {
			(void) printf(
"Test somewhat < %ld regions binary -> F format, depending on precision.\n",
				max_dec_stress_exponent -
				min_dec_stress_exponent + 1);
		}

		if ( test_enabled[(int) etob] || test_enabled[(int) ftob] ||
		     test_enabled[(int) gtob] )
			(void) printf("Test %ld regions decimal -> binary.\n",
				max_bin_stress_exponent -
				min_bin_stress_exponent + 1);
	}

	if ( sigmin == sigmax )
		(void) printf("Test %d decimal digit significance.\n", sigmin );
	else
		(void) printf("Test %d to %d decimal digit significance.\n",
			      sigmin, sigmax);

	if (skip_powers == 0)
		(void) printf(
		    "Test powers of ten and two as well as random data.\n");
	else
		(void) printf("Test only random data.\n");

	if (introduce_errors)
		(void) printf("Introduce an error every %d conversions.\n",
				introduce_errors);

	if (timelimit)
		(void) printf("Each test is limited to %d seconds.\n",
				(int) timelimit);

#if TEST_EXCEPTIONS_DEFAULT == 1
	if (no_exceptions == 1)
		(void) printf("Do not test Exceptions.\n");
#endif

	if (no_stress == 1)
		(void) printf("Do not do stress-testing.\n");

	if (no_sign_zero == 1)
		(void) printf("Do not test signs of zeros.\n");

	(void) printf("\n");
}

int
main(argc, argv)
	int             argc;
	char           *argv[];

{
	init( argc, argv );
	testloop();

	(void) printf( "Performed a total of %lu conversions in %lu regions.\n",
			total_conversions, total_regions );

	(void) printf( "Base test conversion terminates normally.\n");

	return 0;
}
